@@ -1,22 +1,20 @@
 # rawdog: RSS aggregator without delusions of grandeur.
-# Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2012, 2013, 2014, 2015 Adam Sampson <ats@offog.org>
+# Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2012, 2013, 2014, 2015, 2016, 2017, 2018 Adam Sampson <ats@offog.org>
 #
-# rawdog is free software; you can redistribute and/or modify it
-# under the terms of that license as published by the Free Software
-# Foundation; either version 2 of the License, or (at your option)
-# any later version.
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 2 of the License, or
+# (at your option) any later version.
 #
-# rawdog is distributed in the hope that it will be useful, but
-# WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# General Public License for more details.
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with rawdog; see the file COPYING. If not, write to the Free
-# Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-# MA 02110-1301, USA, or see http://www.gnu.org/.
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
-VERSION = "2.21"
+VERSION = "2.23"
 HTTP_AGENT = "rawdog/" + VERSION
 STATE_VERSION = 2
 
@@ -41,6 +39,7 @@ import threading
 import time
 import types
 import urllib2
+import urlparse
 
 try:
 	import tidylib
@@ -52,15 +51,15 @@ try:
 except:
 	mxtidy = None
 
-# Turn off content-cleaning, since we want to see an approximation to the
-# original content for hashing. rawdog will sanitise HTML when writing.
-feedparser.RESOLVE_RELATIVE_URIS = 0
-feedparser.SANITIZE_HTML = 0
-
-# Disable microformat support, because it tends to return poor-quality data
-# (e.g. identifying inappropriate things as enclosures), and it relies on
-# BeautifulSoup which is unable to parse many feeds.
-feedparser.PARSE_MICROFORMATS = 0
+# The sanitisation code was restructured in feedparser 5.3.
+try:
+	_resolveRelativeURIs = feedparser.urls._resolveRelativeURIs
+except AttributeError:
+	_resolveRelativeURIs = feedparser._resolveRelativeURIs
+try:
+	_HTMLSanitizer = feedparser.sanitizer._HTMLSanitizer
+except AttributeError:
+	_HTMLSanitizer = feedparser._HTMLSanitizer
 
 # This is initialised in main().
 persister = None
@@ -118,8 +117,8 @@ def sanitise_html(html, baseurl, inline, config):
 	# "<!doctype html!>"); just remove them all.
 	html = re.sub(r'<![^>]*>', '', html)
 
-	html = feedparser._resolveRelativeURIs(html, baseurl, "UTF-8", type)
-	p = feedparser._HTMLSanitizer("UTF-8", type)
+	html = _resolveRelativeURIs(html, baseurl, "UTF-8", type)
+	p = _HTMLSanitizer("UTF-8", type)
 	p.feed(html)
 	html = p.output()
 
@@ -133,18 +132,29 @@ def sanitise_html(html, baseurl, inline, config):
 			html = "<p>" + html
 
 	if config["tidyhtml"]:
+		# This must include: options where the default value in tidy
+		# has changed at some point, and pytidylib's BASE_OPTIONS which
+		# it would otherwise set if we didn't specify them.
 		args = {
 			"numeric_entities": 1,
+			# In tidy 0.99 these are ASCII; in tidy 5, UTF-8.
+			"input_encoding": "ascii",
+			"output_encoding": "ascii",
 			"output_html": 1,
 			"output_xhtml": 0,
 			"output_xml": 0,
-			"wrap": 0,
+			"indent": 0,
+			"tidy-mark": 0,
+			"alt-text": "",
+			"doctype": "strict",
+			"force-output": 1,
+			# In tidy 0.99, wrap=0 means don't wrap.
+			# In tidy 5, wrap=0 means wrap to width 0.
+			"wrap": 68,
 			}
 		call_hook("mxtidy_args", config, args, baseurl, inline)
 		call_hook("tidy_args", config, args, baseurl, inline)
 		if tidylib is not None:
-			# Disable PyTidyLib's somewhat unhelpful defaults.
-			tidylib.BASE_OPTIONS = {}
 			output = tidylib.tidy_document(html, args)[0]
 		elif mxtidy is not None:
 			output = mxtidy.tidy(html, None, None, **args)[2]
@@ -351,6 +361,12 @@ def is_timeout_exception(exc):
 	"""Return True if the given exception object suggests that a timeout
 	occurred, else return False."""
 
+	if exc is None:
+		return False
+
+	if isinstance(exc, socket.timeout):
+		return True
+
 	# Since urlopen throws away the original exception object,
 	# we have to look at the stringified form to tell if it was a timeout.
 	# (We're in reasonable company here, since test_ssl.py in the Python
@@ -479,18 +495,53 @@ class Feed:
 		if not ":" in url:
 			url = "file:" + url
 
+		parse_args = {
+			"etag": self.etag,
+			"modified": self.modified,
+			"agent": HTTP_AGENT,
+			"handlers": handlers,
+			}
+		# Turn off content-cleaning, as we need the original content
+		# for hashing and we'll do this ourselves afterwards.
+		if hasattr(feedparser, "api"):
+			# feedparser >= 5.3
+			parse_args["sanitize_html"] = False
+			parse_args["resolve_relative_uris"] = False
+		else:
+			# feedparser < 5.3
+			feedparser.RESOLVE_RELATIVE_URIS = 0
+			feedparser.SANITIZE_HTML = 0
+			# Microformat support (removed in 5.3) tends to return
+			# poor-quality data, and relies on BeautifulSoup which
+			# is unable to parse many feeds.
+			feedparser.PARSE_MICROFORMATS = 0
+
 		try:
-			result = feedparser.parse(url,
-				etag=self.etag,
-				modified=self.modified,
-				agent=HTTP_AGENT,
-				handlers=handlers)
+			result = feedparser.parse(url, **parse_args)
+
+			# Older versions of feedparser return some kinds of
+			# download errors in bozo_exception rather than raising
+			# them from feedparser.parse. Normalise this.
+			e = result.get("bozo_exception")
+			if is_timeout_exception(e):
+				result = {"rawdog_timeout": e}
+			elif isinstance(e, urllib2.URLError):
+				result = {"rawdog_exception": e}
 		except Exception, e:
-			result = {
-				"rawdog_exception": e,
-				"rawdog_traceback": sys.exc_info()[2],
-				}
+			if is_timeout_exception(e):
+				result = {"rawdog_timeout": e}
+			else:
+				result = {
+					"rawdog_exception": e,
+					"rawdog_traceback": sys.exc_info()[2],
+					}
 		result["rawdog_responses"] = logger.get_log()
+
+		# For compatibility with old hooks, include an empty "feed"
+		# if a timeout occurred.
+		if "rawdog_timeout" in result:
+			result["feed"] = []
+
 		return result
 
 	def update(self, rawdog, now, config, articles, p):
@@ -522,15 +573,6 @@ class Feed:
 		fatal = False
 		old_url = self.url
 
-		if "rawdog_exception" in p:
-			errors.append("Error fetching or parsing feed:")
-			errors.append(str(p["rawdog_exception"]))
-			if config["showtracebacks"]:
-				from traceback import format_tb
-				errors.append("".join(format_tb(p["rawdog_traceback"])))
-			errors.append("")
-			fatal = True
-
 		if len(responses) != 0 and responses[0]["status"] == 301:
 			# Permanent redirect(s). Find the new location.
 			i = 0
@@ -538,35 +580,47 @@ class Feed:
 				i += 1
 			location = responses[i - 1].get("location")
 
-			if location is None:
+			# According to RFC 2616, the Location header should be
+			# an absolute URI. This doesn't stop the occasional
+			# server sending something like "Location: /" or
+			# "Location: //foo/bar". It's usually a sign of
+			# brokenness, so fail rather than trying to interpret
+			# it liberally.
+			valid_uri = True
+			if location is not None:
+				parsed = urlparse.urlparse(location)
+				if parsed.scheme == "" or parsed.netloc == "":
+					valid_uri = False
+
+			if not valid_uri:
+				errors.append("New URL:     " + location)
+				errors.append("The feed returned a permanent redirect, but with an invalid new location.")
+			elif location is None:
 				errors.append("The feed returned a permanent redirect, but without a new location.")
 			else:
 				errors.append("New URL:     " + location)
 				errors.append("The feed has moved permanently to a new URL.")
 				if config["changeconfig"]:
-					rawdog.change_feed_url(self.url, location, config)
-					errors.append("The config file has been updated automatically.")
+					rawdog.change_feed_url(self.url, location, config, errors.append)
 				else:
 					errors.append("You should update its entry in your config file.")
 			errors.append("")
 
-		bozo_exception = p.get("bozo_exception")
-		got_urlerror = isinstance(bozo_exception, urllib2.URLError)
-		got_timeout = isinstance(bozo_exception, socket.timeout)
-		if got_urlerror or got_timeout:
-			# urllib2 reported an error when fetching the feed.
-			# Check to see if it was a timeout.
-			if not (got_timeout or is_timeout_exception(bozo_exception)):
-				errors.append("Error while fetching feed:")
-				errors.append(str(bozo_exception))
-				errors.append("")
-				fatal = True
-			elif config["ignoretimeouts"]:
+		if "rawdog_timeout" in p:
+			if config["ignoretimeouts"]:
 				return False
 			else:
 				errors.append("Timeout while reading feed.")
 				errors.append("")
 				fatal = True
+		elif "rawdog_exception" in p:
+			errors.append("Error fetching or parsing feed:")
+			errors.append(str(p["rawdog_exception"]))
+			if config["showtracebacks"] and "rawdog_traceback" in p:
+				from traceback import format_tb
+				errors.append("".join(format_tb(p["rawdog_traceback"])))
+			errors.append("")
+			fatal = True
 		elif last_status == 304:
 			# The feed hasn't changed. Return False to indicate
 			# that we shouldn't do expiry.
@@ -1244,13 +1298,13 @@ class Rawdog(Persistable):
 			version = 1
 		return version == STATE_VERSION
 
-	def change_feed_url(self, oldurl, newurl, config):
+	def change_feed_url(self, oldurl, newurl, config, error_fn):
 		"""Change the URL of a feed."""
 
 		assert self.feeds.has_key(oldurl)
 		if self.feeds.has_key(newurl):
-			print >>sys.stderr, "Error: New feed URL is already subscribed; please remove the old one"
-			print >>sys.stderr, "from the config file by hand."
+			error_fn("Error: New feed URL is already subscribed; please remove the old one")
+			error_fn("from the config file by hand.")
 			return
 
 		edit_file("config", ChangeFeedEditor(oldurl, newurl).edit)
@@ -1275,7 +1329,7 @@ class Rawdog(Persistable):
 				if article.feed == oldurl:
 					article.feed = newurl
 
-		print >>sys.stderr, "Feed URL automatically changed."
+		error_fn("The config file has been updated automatically.")
 
 	def list(self, config):
 		"""List the configured feeds."""
@@ -1841,6 +1895,7 @@ Actions (performed in order given):
 
 Special actions (all other options are ignored if one of these is specified):
 --dump URL                   Show what rawdog's parser returns for URL
+--find URL                   Show what rawdog's feed finder returns for URL
 --help                       Display this help and exit
 
 Report bugs to <ats@offog.org>."""
@@ -1862,6 +1917,7 @@ def main(argv):
 			"config=",
 			"dir=",
 			"dump=",
+			"find=",
 			"help",
 			"list",
 			"log=",
@@ -1899,6 +1955,13 @@ def main(argv):
 			import pprint
 			pprint.pprint(feedparser.parse(a, agent=HTTP_AGENT))
 			return 0
+		elif o == "--find":
+			feeds = rawdoglib.feedscanner.feeds(a)
+			if len(feeds) == 0:
+				return 1
+			for url in feeds:
+				print url
+			return 0
 		elif o == "--help":
 			usage()
 			return 0
@@ -1990,4 +2053,3 @@ def main(argv):
 	rawdog_p.close()
 
 	return 0
-
